from django.conf import settings
from django.core.management.base import BaseCommand

from wagtail_airtable.utils import get_all_models


class Command(BaseCommand):
    help = "Looks through every available model in the AIRTABLE_IMPORT_SETTINGS and unsets the `airtable_record_id`"

    def handle(self, *args, **options):
        """
        Gets all the models set in AIRTABLE_IMPORT_SETTINGS, loops through them, and set `airtable_record_id=''` to every one.
        """
        records_updated = 0
        models = get_all_models()
        for model in models:
            if hasattr(model, "airtable_record_id"):
                total_updated = model.objects.update(airtable_record_id="")
                records_updated = records_updated + total_updated

        if options["verbosity"] >= 1:
            self.stdout.write(f"Set {records_updated} objects to airtable_record_id=''")



import sys
from ast import literal_eval
from logging import getLogger

from airtable import Airtable
from django.conf import settings
from django.db import models
from requests import HTTPError

from django.utils.functional import cached_property

from .tests import MockAirtable

logger = getLogger(__name__)


TESTING = any(x in ["test", "runtests.py"] for x in sys.argv)


class AirtableMixin(models.Model):
    """A mixin to update an Airtable when a model object is saved or deleted."""

    AIRTABLE_BASE_KEY = None
    AIRTABLE_TABLE_NAME = None
    AIRTABLE_UNIQUE_IDENTIFIER = None

    # On import, a lot of saving happens, so this attribute gets set to True during import and could be
    # used as a bit of logic to skip a post_save signal, for example.
    _skip_signals = False
    # If the Airtable integration for this model is enabled. Used for sending data to Airtable.
    _is_enabled = False
    # If the Airtable api setup is complete in this model. Used for singleton-like setup_airtable() method.
    _ran_airtable_setup = False
    # Upon save, should this model's data be sent to Airtable?
    # This is an internal variable. Both _push_to_airtable and push_to_airtable needs to be True
    # before a push to Airtable will happen.
    # _push_to_airtable is for internal use only
    _push_to_airtable = False
    # Case for disabling this: when importing data from Airtable as to not
    # ... import data, save the model, and push the same data back to Airtable.
    # push_to_airtable can be set from outside the model
    push_to_airtable = True

    airtable_record_id = models.CharField(max_length=35, db_index=True, blank=True)

    def setup_airtable(self) -> None:
        """
        This method is used in place of __init__() as to not check global settings and
        set the Airtable api client over and over again.

        self._ran_airtable_setup is used to ensure this method is only ever run once.
        """
        if not self._ran_airtable_setup:
            # Don't run this more than once on a model.
            self._ran_airtable_setup = True

            if not hasattr(settings, "AIRTABLE_IMPORT_SETTINGS") or not getattr(
                settings, "WAGTAIL_AIRTABLE_ENABLED", False
            ):
                # No AIRTABLE_IMPORT_SETTINGS were found. Skip checking for settings.
                return None

            # Look for airtable settings. Default to an empty dict.
            AIRTABLE_SETTINGS = settings.AIRTABLE_IMPORT_SETTINGS.get(
                self._meta.label, {}
            )

            self.AIRTABLE_BASE_URL = AIRTABLE_SETTINGS.get("AIRTABLE_BASE_URL", None)
            # Set the airtable settings.
            self.AIRTABLE_BASE_KEY = AIRTABLE_SETTINGS.get("AIRTABLE_BASE_KEY")
            self.AIRTABLE_TABLE_NAME = AIRTABLE_SETTINGS.get("AIRTABLE_TABLE_NAME")
            self.AIRTABLE_UNIQUE_IDENTIFIER = AIRTABLE_SETTINGS.get(
                "AIRTABLE_UNIQUE_IDENTIFIER"
            )
            self.AIRTABLE_SERIALIZER = AIRTABLE_SETTINGS.get("AIRTABLE_SERIALIZER")
            if (
                AIRTABLE_SETTINGS
                and settings.AIRTABLE_API_KEY
                and self.AIRTABLE_BASE_KEY
                and self.AIRTABLE_TABLE_NAME
                and self.AIRTABLE_UNIQUE_IDENTIFIER
            ):
                if not TESTING:
                    self.client = Airtable(
                        self.AIRTABLE_BASE_KEY,
                        self.AIRTABLE_TABLE_NAME,
                        api_key=settings.AIRTABLE_API_KEY,
                    )
                else:
                    self.client = MockAirtable(
                        self.AIRTABLE_BASE_KEY,
                        self.AIRTABLE_TABLE_NAME,
                        api_key=settings.AIRTABLE_API_KEY,
                    )

                self._push_to_airtable = True
                self._is_enabled = True
            else:
                logger.warning(
                    f"Airtable settings are not enabled for the {self._meta.verbose_name} "
                    f"({self._meta.model_name}) model"
                )

    def get_record_usage_url(self):
        if self.is_airtable_enabled and self.AIRTABLE_BASE_URL and self.airtable_record_id:
            url = self.AIRTABLE_BASE_URL.rstrip('/')
            return f"{url}/{self.airtable_record_id}"
        return None

    @property
    def is_airtable_enabled(self):
        """
        Used in the template to determine if a model can or cannot be imported from Airtable.
        """
        if not self._ran_airtable_setup:
            self.setup_airtable()
        return self._is_enabled

    def get_import_fields(self):
        """
        When implemented, should return a dictionary of the mapped fields from Airtable to the model.
        ie.
            {
                "Airtable Column Name": "model_field_name",
                ...
            }
        """
        raise NotImplementedError

    def get_export_fields(self):
        """
        When implemented, should return a dictionary of the mapped fields from Airtable to the model.
        ie.
            {
                "airtable_column": self.airtable_column,
                "annual_fee": self.annual_fee,
            }
        """
        raise NotImplementedError

    @cached_property
    def mapped_export_fields(self):
        return self.get_export_fields()

    def create_record(self) -> dict:
        """
        Create or update a record.

        The create_record() method will look for an Airtable match before trying
        to create a new Airtable record (that comes with a new airtable_record_id).

        This function needs to check for a matched record in Airtable first just in case
        some data became out of sync, or one person worked in Airtable and one worked in
        Wagtail. The idea here is to marry those records whenever possible instead of
        duplicating Airtable records.

        If a record in Airtable exists, update this object with the found record_id. (Prevent record duplication)
        But if a record is NOT found in Airtable, create a new record.
        """
        matched_record = self.match_record()
        if matched_record:
            record = self.update_record(matched_record)
        else:
            record = self.client.insert(self.mapped_export_fields)

        self.airtable_record_id = record["id"]
        return record

    def check_record_exists(self, airtable_record_id) -> bool:
        """
        Check if a record exists in an Airtable by its exact Airtable Record ID.

        This will trigger an Airtable API request.
        Returns a True/False response.
        """
        try:
            record = self.client.get(airtable_record_id)
        except HTTPError:
            record = {}
        return bool(record)

    def update_record(self, airtable_record_id=None):
        """
        Update a record.

        Before updating a record this will check to see if a record even exists
        in Airtable. If a record is not found using its Airtable record_id it cannot
        be updated and may throw an unexpected error.

        If a record DOES exist based on its Airtable record_id, we can update that particular row.
        If a record does NOT exist in Airtable, a new record will need to be created.
        """
        airtable_record_id = airtable_record_id or self.airtable_record_id
        if self.check_record_exists(airtable_record_id):
            # Record exists in Airtable
            record = self.client.update(airtable_record_id, self.mapped_export_fields)
        else:
            # No record exists in Airtable. Create a new record now.
            record = self.create_record()

        self.airtable_record_id = record["id"]
        return record

    def delete_record(self) -> bool:
        """
        Deletes a record from Airtable, but does not delete the object from Django.

        Returns True if the record is successfully deleted, otherwise False.
        """
        try:
            response = self.client.delete(self.airtable_record_id)
            deleted = response["deleted"]
        except HTTPError:
            deleted = False
        return deleted

    def match_record(self) -> str:
        """
        Look for a record in an Airtable. Search by the AIRTABLE_UNIQUE_IDENTIFIER.

        Instead of looking for an Airtable record by it's exact Record ID, it will
        search through the specified Airtable column for a specific value.

        WARNING: If more than one record is found, the first one in the returned
        list of records (a list of dicts) will be used.

        This differs from check_record_exists() as this will return the record string
        (or an empty string if a record is not found), whereas check_record_exists()
        will return a True/False boolean to let you know if a record simply exists,
        or doesn't exist.
        """
        if type(self.AIRTABLE_UNIQUE_IDENTIFIER) == dict:
            keys = list(self.AIRTABLE_UNIQUE_IDENTIFIER.keys())
            values = list(self.AIRTABLE_UNIQUE_IDENTIFIER.values())
            # TODO: Edge case handling:
            #       - Handle multiple dictionary keys
            #       - Handle empty dictionary
            airtable_column_name = keys[0]
            model_field_name = values[0]
            value = getattr(self, model_field_name)
        else:
            _airtable_unique_identifier = self.AIRTABLE_UNIQUE_IDENTIFIER
            value = getattr(self, _airtable_unique_identifier)
            airtable_column_name = self.AIRTABLE_UNIQUE_IDENTIFIER
        records = self.client.search(airtable_column_name, value)
        total_records = len(records)
        if total_records:
            # If more than 1 record was returned log a warning.
            if total_records > 1:
                logger.info(
                    f"Found {total_records} Airtable records for {airtable_column_name}={value}. "
                    f"Using first available record ({records[0]['id']}) and ignoring the others."
                )
            # Always return the first record
            return records[0]["id"]

        return ""

    def refresh_mapped_export_fields(self) -> None:
        """Delete the @cached_property caching on self.mapped_export_fields."""
        try:
            del self.mapped_export_fields
        except Exception:
            # Doesn't matter what the error is.
            pass

    @classmethod
    def parse_request_error(cls, error):
        """
        Parse an Airtable/requests HTTPError string.

        Example: 401 Client Error: Unauthorized for url: https://api.airtable.com/v0/appYourAppId/Your%20Table?filterByFormula=.... [Error: {'type': 'AUTHENTICATION_REQUIRED', 'message': 'Authentication required'}]
        """
        if not error:
            return {
                "status_code": 503,
                "type": "SERVICE_UNAVAILABLE",
                "message": "Service may be down, or is otherwise uncreachable"
            }

        code = int(error.split(":", 1)[0].split(" ")[0])
        if code == 502:
            return {
                "status_code": code,
                "type": "SERVER_ERROR",
                "message": "Service may be down, or is otherwise uncreachable"
            }        
        error_json = error.split("[Error: ")[1].rstrip("]")
        if error_json == "NOT_FOUND":  # 404's act different
            return {
                "status_code": code,
                "type": "NOT_FOUND",
                "message": "Record not found",
            }
        else:
            error_info = literal_eval(error_json)
            return {
                "status_code": code,
                "type": error_info["type"],
                "message": error_info["message"],
            }

    def save(self, *args, **kwargs):
        """
        If there's an existing airtable record id, update the row.
        Otherwise attempt to create a new record.
        """
        saved_model = super().save(*args, **kwargs) # Save to database first so we get pk, in case it's used for uniqueness
        self.setup_airtable()
        if self._push_to_airtable and self.push_to_airtable:
            # Every airtable model needs mapped fields.
            # mapped_export_fields is a cached property. Delete the cached prop and get new values upon save.
            self.refresh_mapped_export_fields()
            if self.airtable_record_id:
                # If this model has an airtable_record_id, attempt to update the record.
                try:
                    self.update_record()
                except HTTPError as e:
                    error = self.parse_request_error(e.args[0])
                    message = (
                        f"Could not update Airtable record. Reason: {error['message']}"
                    )
                    logger.warning(message)
                    # Used in the `after_edit_page` hook. If it exists, an error message will be displayed.
                    self._airtable_update_error = message
            else:
                # Creating a record will also search for an existing field match
                # ie. Looks for a matching `slug` in Airtable and Wagtail/Django
                try:
                    self.create_record()
                    # Save once more so the airtable_record_id is stored.
                    super().save(*args, **kwargs)
                except HTTPError as e:
                    error = self.parse_request_error(e.args[0])
                    message = (
                        f"Could not create Airtable record. Reason: {error['message']}"
                    )
                    logger.warning(message)
                    # Used in the `after_edit_page` hook. If it exists, an error message will be displayed.
                    self._airtable_update_error = message

        return saved_model

    def delete(self, *args, **kwargs):
        self.setup_airtable()
        if self.push_to_airtable and self._push_to_airtable and self.airtable_record_id:
            # Try to delete the record from the Airtable.
            self.delete_record()
        return super().delete(*args, **kwargs)

    class Meta:
        abstract = True



from django.conf import settings
from django.conf.urls import url
from django.urls import reverse
from wagtail.core import hooks
from wagtail.admin.menu import MenuItem

from wagtail_airtable.views import AirtableImportListing
from wagtail_airtable.utils import airtable_message, can_send_airtable_messages
from .mixins import AirtableMixin


@hooks.register("register_admin_urls")
def register_airtable_url():
    return [
        url(
            r"^airtable-import/$",
            AirtableImportListing.as_view(),
            name="airtable_import_listing",
        ),
    ]


@hooks.register("register_settings_menu_item")
def register_airtable_setting():
    def is_shown(request):
        return getattr(settings, "WAGTAIL_AIRTABLE_ENABLED", False)

    menu_item = MenuItem(
        "Airtable Import",
        reverse("airtable_import_listing"),
        classnames="icon icon-cog",
        order=1000,
    )
    menu_item.is_shown = is_shown
    return menu_item


@hooks.register("after_edit_page")
def after_page_update(request, page):
    if can_send_airtable_messages(page):
        airtable_message(request, page)


@hooks.register("after_create_snippet")
@hooks.register("after_edit_snippet")
def after_snippet_update(request, instance):
    if can_send_airtable_messages(instance):
        airtable_message(request, instance)


@hooks.register("after_delete_snippet")
def after_snippet_delete(request, instances):
    total_deleted = len(instances)
    instance = instances[0]
    if can_send_airtable_messages(instance):
        message = f"Airtable record deleted"
        if total_deleted > 1:
            message = f"{total_deleted} Airtable records deleted"
        airtable_message(request, instance, message=message, buttons_enabled=False)



# Generated by Django 3.0.6 on 2020-05-15 18:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ModelNotUsed',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('airtable_record_id', models.CharField(blank=True, db_index=True, max_length=35)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



"""A mocked Airtable API wrapper."""
from unittest import mock


class MockAirtable(mock.Mock):
    pass


MockAirtable.table_name = "app_airtable_advert_base_key"

MockAirtable.get = mock.MagicMock("get")
MockAirtable.get.return_value = {
    "id": "recNewRecordId",
    "fields": {
        "title": "Red! It's the new blue!",
        "description": "Red is a scientifically proven color that moves faster than all other colors.",
        "external_link": "https://example.com/",
        "is_active": True,
        "rating": "1.5",
        "long_description": "<p>Lorem ipsum dolor sit amet, consectetur adipisicing elit. Veniam laboriosam consequatur saepe. Repellat itaque dolores neque, impedit reprehenderit eum culpa voluptates harum sapiente nesciunt ratione.</p>",
        "points": 95,
        "slug": "red-its-new-blue",
    },
}

MockAirtable.insert = mock.MagicMock("insert")
MockAirtable.insert.return_value = {
    "id": "recNewRecordId",
    "fields": {
        "title": "Red! It's the new blue!",
        "description": "Red is a scientifically proven color that moves faster than all other colors.",
        "external_link": "https://example.com/",
        "is_active": True,
        "rating": "1.5",
        "long_description": "<p>Lorem ipsum dolor sit amet, consectetur adipisicing elit. Veniam laboriosam consequatur saepe. Repellat itaque dolores neque, impedit reprehenderit eum culpa voluptates harum sapiente nesciunt ratione.</p>",
        "points": 95,
        "slug": "red-its-new-blue",
    },
}

MockAirtable.update = mock.MagicMock("update")
MockAirtable.update.return_value = {
    "id": "recNewRecordId",
    "fields": {
        "title": "Red! It's the new blue!",
        "description": "Red is a scientifically proven color that moves faster than all other colors.",
        "external_link": "https://example.com/",
        "is_active": True,
        "rating": "1.5",
        "long_description": "<p>Lorem ipsum dolor sit amet, consectetur adipisicing elit. Veniam laboriosam consequatur saepe. Repellat itaque dolores neque, impedit reprehenderit eum culpa voluptates harum sapiente nesciunt ratione.</p>",
        "points": 95,
        "slug": "red-its-new-blue",
    },
}

MockAirtable.delete = mock.MagicMock("delete")
MockAirtable.delete.return_value = {"deleted": True, "record": "recNewRecordId"}

MockAirtable.search = mock.MagicMock("search")
MockAirtable.search.return_value = [
    {
        "id": "recNewRecordId",
        "fields": {
            "title": "Red! It's the new blue!",
            "description": "Red is a scientifically proven color that moves faster than all other colors.",
            "external_link": "https://example.com/",
            "is_active": True,
            "rating": "1.5",
            "long_description": "<p>Lorem ipsum dolor sit amet, consectetur adipisicing elit. Veniam laboriosam consequatur saepe. Repellat itaque dolores neque, impedit reprehenderit eum culpa voluptates harum sapiente nesciunt ratione.</p>",
            "points": 95,
            "slug": "red-its-new-blue",
        },
    },
    {
        "id": "Different record",
        "fields": {
            "title": "Not the used record.",
            "description": "This is only used for multiple responses from MockAirtable",
            "external_link": "https://example.com/",
            "is_active": False,
            "rating": "5.5",
            "long_description": "",
            "points": 1,
            "slug": "not-the-used-record",
        },
    },
]

MockAirtable.get_all = mock.MagicMock("get_all")
MockAirtable.get_all.return_value = [
    {
        "id": "recNewRecordId",
        "fields": {
            "title": "Red! It's the new blue!",
            "description": "Red is a scientifically proven color that moves faster than all other colors.",
            "external_link": "https://example.com/",
            "is_active": True,
            "rating": "1.5",
            "long_description": "<p>Lorem ipsum dolor sit amet, consectetur adipisicing elit. Veniam laboriosam consequatur saepe. Repellat itaque dolores neque, impedit reprehenderit eum culpa voluptates harum sapiente nesciunt ratione.</p>",
            "points": 95,
            "slug": "red-its-new-blue",
        },
    },
    {
        "id": "Different record",
        "fields": {
            "title": "Not the used record.",
            "description": "This is only used for multiple responses from MockAirtable",
            "external_link": "https://example.com/",
            "is_active": False,
            "rating": "5.5",
            "long_description": "",
            "points": 1,
            "slug": "not-the-used-record",
        },
    },
    {
        "id": "recRecordThree",
        "fields": {
            "title": "A third record.",
            "description": "This is only used for multiple responses from MockAirtable",
            "external_link": "https://example.com/",
            "is_active": False,
            "rating": "5.5",
            "long_description": "",
            "points": 1,
            "slug": "record-3",
        },
    },
    {
        "id": "recRecordFour",
        "fields": {
            "title": "A fourth record.",
            "description": "This is only used for multiple responses from MockAirtable",
            "external_link": "https://example.com/",
            "is_active": False,
            "rating": "5.5",
            "long_description": "",
            "points": 1,
            "slug": "record-4",
        },
    },
]



# Generated by Django 3.0.6 on 2020-05-15 18:16

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0002_modelnotused'),
    ]

    operations = [
        migrations.CreateModel(
            name='SimilarToAdvert',
            fields=[
                ('advert_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='tests.Advert')),
            ],
            options={
                'abstract': False,
            },
            bases=('tests.advert',),
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-14 19:00

from django.db import migrations, models
import django.db.models.deletion
import wagtail.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0045_assign_unlock_grouppagepermission'),
    ]

    operations = [
        migrations.CreateModel(
            name='Advert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('airtable_record_id', models.CharField(blank=True, db_index=True, max_length=35)),
                ('title', models.CharField(max_length=255)),
                ('description', models.TextField(blank=True)),
                ('external_link', models.URLField(blank=True, max_length=500)),
                ('is_active', models.BooleanField(default=False)),
                ('rating', models.DecimalField(choices=[(1.0, '1'), (1.5, '1.5'), (2.0, '2'), (2.5, '2.5'), (3.0, '3'), (3.5, '3.5'), (4.0, '4'), (4.5, '4.5'), (5.0, '5')], decimal_places=1, max_digits=2, null=True)),
                ('long_description', wagtail.core.fields.RichTextField(blank=True, null=True)),
                ('points', models.IntegerField(blank=True, null=True)),
                ('slug', models.SlugField(max_length=100, unique=True)),
            ],
            options={
                'verbose_name': 'Advertisement',
                'verbose_name_plural': 'Advertisements',
            },
        ),
        migrations.CreateModel(
            name='SimplePage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('intro', models.TextField()),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
    ]



from django.conf import settings
from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect
from django.urls import reverse
from django.views.generic import TemplateView
from django.core.exceptions import ObjectDoesNotExist
from logging import getLogger

from wagtail_airtable.forms import AirtableImportModelForm
from wagtail_airtable.utils import get_model_for_path

from wagtail_airtable.management.commands.import_airtable import Importer

logger = getLogger(__name__)


class AirtableImportListing(TemplateView):
    """
    Loads options for importing Airtable data
    """

    template_name = "wagtail_airtable/airtable_import_listing.html"
    http_method_names = ["get", "post"]

    def post(self, request, *args, **kwargs):
        form = AirtableImportModelForm(request.POST)
        if form.is_valid():
            model_label = form.cleaned_data["model"]
            importer = Importer(models=[model_label], options={"verbosity": 1})
            importer.run()
            message = f"{importer.created} items created. {importer.updated} items updated. {importer.skipped} items skipped."
            messages.add_message(
                request, messages.SUCCESS, f"Import succeeded with {message}"
            )
        else:
            messages.add_message(request, messages.ERROR, "Could not import")

        return redirect(reverse("airtable_import_listing"))

    def _get_base_model(self, model):
        """
        For the given model, return the highest concrete model in the inheritance tree -
        e.g. for BlogPage, return Page
        """
        if model._meta.parents:
            model = model._meta.get_parent_list()[0]
        return model

    def get_validated_models(self):
        """Get models from AIRTABLE_IMPORT_SETTINGS, validate they exist, and return a list of tuples.

        returns:
            [
                ('Credit Card', 'creditcards.CreditCard', <CreditCard: Cardname.=>),
                ('..', '..'),
            ]
        """
        airtable_settings = getattr(settings, "AIRTABLE_IMPORT_SETTINGS", {})

        # Loop through all the models in the settings and create a new dict
        # of the unique settings for each model label.
        # If settings were used more than once the second (3rd, 4th, etc) common settings
        # will be bulked into a "grouped_models" list.
        tracked_settings = []
        models = {}
        for label, model_settings in airtable_settings.items():
            if model_settings not in tracked_settings:
                tracked_settings.append(model_settings)
                models[label] = model_settings
                models[label]["grouped_models"] = []
            else:
                for label2, model_settings2 in models.items():
                    if model_settings is model_settings2:
                        models[label2]["grouped_models"].append(label)

        # Validated models are models that actually exist.
        # This way fake models can't be added.
        validated_models = []
        for label, model_settings in models.items():
            # If this model is allowed to be imported. Default is True.
            if model_settings.get("AIRTABLE_IMPORT_ALLOWED", True):
                # A temporary variable for holding grouped model names.
                # This is added to the validated_models item later.
                # This is only used for displaying model names in the import template
                _grouped_models = []
                # Loop through the grouped_models list in each setting, validate each model,
                # then add it to the larger grouped_models
                if model_settings.get("grouped_models"):
                    for grouped_model_label in model_settings.get("grouped_models"):
                        if "." in grouped_model_label:
                            model = get_model_for_path(grouped_model_label)
                            if model:
                                _grouped_models.append(model._meta.verbose_name_plural)

                if "." in label:
                    model = get_model_for_path(label)
                    if model:
                        # Append a triple-tuple to the validated_models with the:
                        # (1. Models verbose name, 2. Model label, 3. is_airtable_enabled from the model, and 4. List of grouped models)
                        airtable_enabled_for_model = getattr(
                            model, "is_airtable_enabled", False
                        )
                        validated_models.append(
                            (
                                model._meta.verbose_name_plural,
                                label,
                                airtable_enabled_for_model,
                                _grouped_models,
                            )
                        )
                    else:
                        raise ImproperlyConfigured(
                            "%r is not recognised as a model name." % label
                        )

        return validated_models

    def get_context_data(self, **kwargs):
        """Add validated models from the AIRTABLE_IMPORT_SETTINGS to the context."""
        return {"models": self.get_validated_models()}



from django import template
from django.conf import settings


register = template.Library()

@register.simple_tag
def can_import_model(model_label) -> bool:
    """
    Check if a model can be imported based on its model label.

    Use:
    {% load wagtail_airtable_tags %}
    {% can_import_model "yourapp.ModelName" as template_var %}

    Returns True or False.
    """
    airtable_settings = getattr(settings, "AIRTABLE_IMPORT_SETTINGS", {})
    has_settings = airtable_settings.get(model_label, False)
    return bool(has_settings)



import sys
from importlib import import_module

from airtable import Airtable
from django.db import models, IntegrityError
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand
from logging import getLogger
from modelcluster.contrib.taggit import ClusterTaggableManager
from taggit.managers import TaggableManager
from wagtail.core.models import Page

from wagtail_airtable.tests import MockAirtable
from wagtail_airtable.utils import get_validated_models

logger = getLogger(__name__)


DEFAULT_OPTIONS = {
    "verbosity": 1,
}

TESTING = any(x in ["test", "runtests.py"] for x in sys.argv)


class Importer:
    def __init__(self, models=[], options=DEFAULT_OPTIONS):
        self.models = models
        self.options = options
        self.records_used = []
        self.cached_records = {}
        self.created = 0
        self.updated = 0
        self.skipped = 0

    def debug_message(self, message):
        """
        Local function. Print debug messages if `verbosity` is 2 or higher.
        """
        if self.options["verbosity"] >= 2:
            if not TESTING:
                print(message)
            return message

    def get_model_serializer(self, serializer_string):
        location, serializer_name = serializer_string.rsplit(".", 1)
        module = import_module(location)
        serializer_class = getattr(module, serializer_name)
        return serializer_class

    def get_model_settings(self, model) -> dict:
        return settings.AIRTABLE_IMPORT_SETTINGS.get(model._meta.label, {})

    def get_column_to_field_names(self, airtable_unique_identifier) -> tuple:
        uniq_id_type = type(airtable_unique_identifier)
        airtable_unique_identifier_column_name = None
        airtable_unique_identifier_field_name = None
        if uniq_id_type == str:
            # The unique identifier is a string.
            # Use it as the Airtable Column name and the Django field name
            airtable_unique_identifier_column_name = airtable_unique_identifier
            airtable_unique_identifier_field_name = airtable_unique_identifier
        elif uniq_id_type == dict:
            # Unique identifier is a dictionary.
            # Use the key as the Airtable Column name and the value as the Django Field name.
            (
                airtable_unique_identifier_column_name,
                airtable_unique_identifier_field_name,
            ) = list(airtable_unique_identifier.items())[0]

        return (
            airtable_unique_identifier_column_name,
            airtable_unique_identifier_field_name,
        )

    def get_or_set_cached_records(self, airtable_client):
        # Memoize results from Airtable so we don't hit the same API multiple times
        # This is largely used to support additional Wagtail/Airatble settings
        # that are identical to each other, as in they would use the
        # same Airtable Base as the key in the dictionary.
        #   ie.
        #   'yourapp.YourPage': {
        #       ...
        #       'AIRTABLE_TABLE_NAME': 'Your Table',
        #   },
        #   'different_app.DifferentPage': {
        #       ...
        #       'AIRTABLE_TABLE_NAME': 'Your Table',  # Same Airtable Table name
        #   }
        #   All of the above settings will use the 'Your Table' results
        #   instead of hitting the Airtable API for each model and getting the same
        #   results every time. This is designed to help with API efficiency, reduce
        #   load/import times, and to reduce how much memory is required to save all
        #   the records from Airtable.
        if self.cached_records.get(airtable_client.table_name):
            all_records = self.cached_records.get(airtable_client.table_name)
        else:
            # Get all the airtable records for the specified table.
            all_records = airtable_client.get_all()
            self.cached_records[airtable_client.table_name] = all_records
        return all_records

    def convert_mapped_fields(self, record_fields_dict, mapped_fields_dict) -> dict:
        # Create a dictionary of newly mapped key:value pairs based on the `mappings` dict above.
        # This wil convert "airtable column name" to "django_field_name"
        mapped_fields_dict = {
            mapped_fields_dict[key]: value
            for (key, value) in record_fields_dict.items()
            if key in mapped_fields_dict
        }
        return mapped_fields_dict

    def update_object(
        self, instance, record_id, serialized_data, is_wagtail_model=False
    ) -> bool:
        """
        Attempts to update an object.

        Returns a bool that determines if the object was updated or not.
        """

        if serialized_data.is_valid():
            self.debug_message(
                "\t\t Serializer data was valid. Setting attrs on model..."
            )
            model = type(instance)
            for field_name, value in serialized_data.validated_data.items():
                field_type = type(
                    model._meta.get_field(field_name)
                )  # ie. django.db.models.fields.CharField
                # If this field type is a subclass of a known Wagtail Tag, or a Django m2m field
                # We need to loop through all the values and add them to the m2m-style field.
                if issubclass(
                    field_type,
                    (TaggableManager, ClusterTaggableManager, models.ManyToManyField,),
                ):
                    m2m_field = getattr(instance, field_name)
                    for m2m_value in value:
                        m2m_field.add(m2m_value)
                else:
                    setattr(instance, field_name, value)
            # When an object is saved it should NOT push its newly saved data back to Airtable.
            # This could theoretically cause a loop. By default this setting is True. But the
            # below line confirms it's false, just to be safe.
            instance.airtable_record_id = record_id
            instance.push_to_airtable = False
            try:
                if is_wagtail_model:
                    self.debug_message("\t\t This is a Wagtail Page model")
                    # Wagtail page. Requires a .save_revision()
                    if not instance.locked:
                        self.debug_message(
                            "\t\t\t Page is not locked. Saving page and creating a new revision."
                        )
                        # Only save the page if the page is not locked
                        instance._skip_signals=True
                        instance.save()
                        instance.save_revision()
                        self.updated = self.updated + 1
                    else:
                        self.debug_message("\t\t\t Page IS locked. Skipping Page save.")
                        self.skipped = self.skipped + 1
                else:
                    # Django model. Save normally.
                    self.debug_message("\t\t Saving Django model")
                    instance._skip_signals=True
                    instance.save()
                    self.updated = self.updated + 1

                # New record being processed. Save it to the list of records.
                self.records_used.append(record_id)
                # Object updated (and record was used)
                return True
            except ValidationError as error:
                self.skipped = self.skipped + 1
                error_message = "; ".join(error.messages)
                logger.error(
                    f"Unable to save {instance._meta.label} -> '{instance}'. Error(s): {error_message}"
                )
                self.debug_message(
                    f"\t\t Could not save Wagtail/Django model. Error: {error_message}"
                )
        else:
            logger.info(f"Invalid data for record {record_id}")
            self.debug_message(
                f"\t\t Serializer was invalid for record: {record_id}, model id: {instance.pk}"
            )
            self.debug_message(
                "\t\t Continuing to look for object by its unique identifier"
            )
        # Not updated.
        return False

    def update_object_by_uniq_col_name(
        self,
        field_mapping=None,
        model=None,
        serialized_data=None,
        record_id=None,
        is_wagtail_model=False,
    ):
        k, v = zip(*field_mapping.items())
        airtable_unique_identifier_field_name = k[0]
        unique_identifier = v[0]

        if unique_identifier:
            self.debug_message(
                f"\t\t An Airtable record based on the unique identifier was found: {airtable_unique_identifier_field_name}"
            )
            try:
                instance = model.objects.get(
                    **{airtable_unique_identifier_field_name: unique_identifier}
                )
                self.debug_message(
                    f"\t\t Local object found by Airtable unique column name: {airtable_unique_identifier_field_name}"
                )
            except model.DoesNotExist:
                instance = None
                self.debug_message(
                    "\t\t No object was found based on the Airtable column name"
                )

            if instance:
                # A local model object was found by a unique identifier.
                if serialized_data.is_valid():
                    for field_name, value in serialized_data.validated_data.items():
                        field_type = type(
                            model._meta.get_field(field_name)
                        )  # ie. django.db.models.fields.CharField
                        # If this field type is a subclass of a known Wagtail Tag, or a Django m2m field
                        # We need to loop through all the values and add them to the m2m-style field.
                        if issubclass(
                            field_type,
                            (
                                TaggableManager,
                                ClusterTaggableManager,
                                models.ManyToManyField,
                            ),
                        ):
                            m2m_field = getattr(instance, field_name)
                            for m2m_value in value:
                                m2m_field.add(m2m_value)
                        else:
                            setattr(instance, field_name, value)
                    # When an object is saved it should NOT push its newly saved data back to Airtable.
                    # This could theoretically cause a loop. By default this setting is False. But the
                    # below line confirms it's false, just to be safe.
                    instance.airtable_record_id = record_id
                    instance.push_to_airtable = False
                    try:
                        if is_wagtail_model:
                            # Wagtail page. Requires a .save_revision()
                            if not instance.locked:
                                # Only save the page if the page is not locked
                                instance._skip_signals=True
                                instance.save()
                                instance.save_revision()
                                self.updated = self.updated + 1
                            else:
                                self.debug_message(
                                    "\t\t\t Page IS locked. Skipping Page save."
                                )
                                self.skipped = self.skipped + 1
                        else:
                            # Django model. Save normally.
                            instance._skip_signals=True
                            instance.save()
                            self.debug_message("\t\t\t Saved!")
                            self.updated = self.updated + 1

                        # Record this record as "used"
                        self.records_used.append(record_id)
                        # New record being processed. Save it to the list of records.
                        return True
                    except ValidationError as error:
                        error_message = "; ".join(error.messages)
                        logger.error(
                            f"Unable to save {instance}. Error(s): {error_message}"
                        )
                        self.debug_message(
                            f"\t\t Unable to save {instance} (ID: {instance.pk}; Airtable Record ID: {record_id}). Reason: {error_message}"
                        )
                        self.skipped = self.skipped + 1
                else:
                    logger.info(f"Invalid data for record {record_id}")
                    self.debug_message("\t\t Serializer data was invalid.")
                    self.skipped = self.skipped + 1
            else:
                # No object was found by this unique ID.
                # Do nothing. The next step will be to create this object in Django
                logger.info(
                    f"{model._meta.verbose_name} with field {airtable_unique_identifier_field_name}={unique_identifier} was not found"
                )
                self.debug_message(
                    f"\t\t {model._meta.verbose_name} with field {airtable_unique_identifier_field_name}={unique_identifier} was not found"
                )
        else:
            # There was no unique identifier set for this model.
            # Nothing can be done about that right now.
            logger.info(f"{model._meta.verbose_name} does not have a unique identifier")
            self.debug_message(
                f"\t\t {model._meta.verbose_name} does not have a unique identifier"
            )
            self.skipped = self.skipped + 1
        return False

    def is_wagtail_page(self, model):
        if issubclass(model, Page):
            return True
        return False

    def get_data_for_new_model(self, serialized_data, mapped_import_fields, record_id):

        # Check if we can use the serialized data to create a new model.
        # If we can, great! If not, fall back to the original mapped_import_fields
        # If this has to fall back to the original mapped_import_fields: failure
        # to create a model will be higher than normal.
        if serialized_data.is_valid():
            data_for_new_model = dict(serialized_data.validated_data)
        else:
            data_for_new_model = mapped_import_fields
        data_for_new_model["airtable_record_id"] = record_id

        # First things first, remove any "pk" or "id" items from the mapped_import_fields
        # This will let Django and Wagtail handle the PK on its own, as it should.
        # When the model is saved it'll trigger a push to Airtable and automatically update
        # the necessary column with the new PK so it's always accurate.
        for key in (
            "pk",
            "id",
        ):
            try:
                del data_for_new_model[key]
            except KeyError:
                pass

        return data_for_new_model

    def run(self):

        models = get_validated_models(models=self.models)
        self.debug_message(f"Validated models: {models}")

        # Used for re-using API data instead of making several of API request and waiting/spamming the Airtable API
        # Maintain a list of record Ids that were used already. Every record is a unique ID so processing the
        # Same record more than once will just be hitting the DB over and over and over again. No bueno.

        for model in models:
            self.debug_message(f"IMPORTING MODEL: {model}")
            # Wagtail models require a .save_revision() call when being saved.
            is_wagtail_model = issubclass(model, Page)
            # Airtable global settings.
            airtable_settings = self.get_model_settings(model)

            # Set the unique identifier and serializer.
            model_serializer = self.get_model_serializer(
                airtable_settings.get("AIRTABLE_SERIALIZER")
            )
            # Get the unique column name and field name.
            # The CAN be the same value if a string is provided in the settings.
            (
                airtable_unique_identifier_column_name,
                airtable_unique_identifier_field_name,
            ) = self.get_column_to_field_names(
                airtable_settings.get("AIRTABLE_UNIQUE_IDENTIFIER")
            )

            if (
                not airtable_unique_identifier_field_name
                and not airtable_unique_identifier_column_name
            ):
                logger.error("No unique columns are set in your Airtable configuration")
                continue

            # Set the Airtable API client on a per-model basis
            if not TESTING:
                airtable = Airtable(
                    airtable_settings.get("AIRTABLE_BASE_KEY"),
                    airtable_settings.get("AIRTABLE_TABLE_NAME"),
                    api_key=settings.AIRTABLE_API_KEY,
                )
            else:
                airtable = MockAirtable(
                    airtable_settings.get("AIRTABLE_BASE_KEY"),
                    airtable_settings.get("AIRTABLE_TABLE_NAME"),
                    api_key=settings.AIRTABLE_API_KEY,
                )

            all_records = self.get_or_set_cached_records(airtable)

            self.debug_message(
                f"\t Airtable base key: {airtable_settings.get('AIRTABLE_BASE_KEY')}"
            )
            self.debug_message(
                f"\t Airtable table name: {airtable_settings.get('AIRTABLE_TABLE_NAME')}"
            )
            self.debug_message("\t Airtable unique identifier settings:")
            self.debug_message(f"\t Airtable records: {len(all_records)}")
            self.debug_message(
                f"\t\t Airtable column: {airtable_unique_identifier_column_name}"
            )
            self.debug_message(
                f"\t\t Django field name: {airtable_unique_identifier_field_name}"
            )

            # Loop through every record in the Airtable.
            for record in all_records:
                # If a record was used already, skip this iteration.
                if record["id"] in self.records_used:
                    continue


                record_id = record["id"]
                record_fields = record["fields"]
                mapped_import_fields = self.convert_mapped_fields(
                    record_fields, model.map_import_fields()
                )
                serialized_data = model_serializer(data=mapped_import_fields)
                serialized_data.is_valid()

                # Look for a record by its airtable_record_id.
                # If it exists, update the data.
                self.debug_message(
                    f"\n\t Looking for existing object with record: {record_id}"
                )
                try:
                    obj = model.objects.get(airtable_record_id=record_id)
                    self.debug_message(f"\t\t Local object found {obj}")
                except model.DoesNotExist:
                    obj = None
                    self.debug_message("\t\t Local object was NOT found")
                except model.MultipleObjectsReturned:
                    # In the case that multiple models have gotten the same record_id
                    objs = model.objects.filter(airtable_record_id=record_id)
                    # Set the first one to be the "right" one
                    obj = objs[0]
                    for ob in objs[1:]:
                        # set airtable_record_id on the "impostors" to ""
                        ob.airtable_record_id = ""
                        ob.save()


                if obj:
                    # Model object was found by it's airtable_record_id
                    was_updated = self.update_object(
                        instance=obj,
                        record_id=record_id,
                        serialized_data=serialized_data,
                        is_wagtail_model=is_wagtail_model,
                    )
                    # Object was updated. No need to continue through the rest of this function
                    if was_updated:
                        continue

                # This `unique_identifier` is the value of an Airtable record.
                # ie.
                #   fields = {
                #         'Slug': 'your-model'
                #   }
                # This will return 'your-model' and can now be searched for as model.objects.get(slug='your-model')
                unique_identifier = record_fields.get(
                    airtable_unique_identifier_column_name, None
                )
                was_updated = self.update_object_by_uniq_col_name(
                    field_mapping={
                        airtable_unique_identifier_field_name: unique_identifier
                    },
                    model=model,
                    serialized_data=serialized_data,
                    record_id=record_id,
                    is_wagtail_model=is_wagtail_model,
                )
                if was_updated:
                    continue

                # Look for the a PARENT_PAGE_ID setting and hopefully find a corresponding
                # Wagtail Page to set as the Parent Page. If the `parent_page` is set in
                # this conditional, it will be used later to create a new page.
                parent_page = None
                if self.is_wagtail_page(model):
                    # Page record was used. This is for caching duplicate entries to avoid too many API calls.
                    self.records_used.append(record_id)
                    self.debug_message("\t\t Checking PARENT_PAGE_ID settings...")

                    # Check if the PARENT_PAGE_ID setting is set. If it's not set in the Django settings,
                    # Don't try to create a new Wagtail Page.
                    parent_page_id_setting = airtable_settings.get("PARENT_PAGE_ID", None)
                    if parent_page_id_setting:
                        self.debug_message("\t\t Setting found...")

                        # Check if the PARENT_PAGE_ID is a callable, string location to a function, or an int ID.
                        if callable(parent_page_id_setting):
                            # A function was passed into the settings. Execute it.
                            parent_page_id = parent_page_id_setting()
                        elif isinstance(parent_page_id_setting, str):
                            # A string location was passed into the settings
                            # Parse the string function location and execute it
                            location, function_name = parent_page_id_setting.rsplit(".", 1)
                            module = import_module(location)
                            parent_page_callable = getattr(module, function_name)
                            parent_page_id = parent_page_callable()
                        else:
                            # This should be an integer representing the Parent Page to nest pages under
                            parent_page_id = parent_page_id_setting

                        try:
                            # Look for the parent page ID. If it doesn't exist, a page cannot be created.
                            # This needs to be set to a value for later use.
                            parent_page = Page.objects.get(pk=parent_page_id)
                        except Page.DoesNotExist:
                            logger.info(
                                f"Could not create new Page object. Parent page with ID {parent_page_id} doesn't exist"
                            )
                            self.debug_message(
                                f"\t\t Could not create new Page object. Parent page with ID {parent_page_id} doesn't exist"
                            )
                            continue

                # Attempt to format valid data to create a new model form either the
                # validated data in the serializer, or the mapped_field data.
                data_for_new_model = self.get_data_for_new_model(
                    serialized_data, mapped_import_fields, record_id
                )

                # If there is no match whatsoever, try to create a new `model` instance.
                # Note: this may fail if there isn't enough data in the Airtable record.
                try:
                    self.debug_message("\t\t Attempting to create a new Page...")
                    new_model = model(**data_for_new_model)
                    new_model._skip_signals=True
                    # `parent_page` is either None or a Page instance.
                    if parent_page:
                        # Pages are not live by default, and always have unpublished changes.
                        new_model.live = False
                        new_model.has_unpublished_changes = True

                        # Add the new page to the parent page.
                        parent_page.add_child(instance=new_model)

                        # If the settings are set to auto publish a page when it's imported,
                        # save the page revision and publish it. Otherwise just save the revision.
                        if airtable_settings.get("AUTO_PUBLISH_NEW_PAGES", False):
                            self.debug_message("\t\t Publishing new Page...")
                            new_model.save_revision().publish()
                        else:
                            new_model.save_revision()
                        self.debug_message("\t\t Page created")
                    else:
                        new_model.save()
                        self.debug_message("\t\t Object created")
                    self.created = self.created + 1
                except ValueError as value_error:
                    logger.info(
                        f"Could not create new model object. Value Error: {value_error}"
                    )
                    self.debug_message(
                        f"\t\t Could not create new model object. Value Error: {value_error}"
                    )
                except IntegrityError as e:
                    logger.info(
                        f"Could not create new model object. Integrity Error: {e}"
                    )
                    self.debug_message(
                        f"\t\t Could not create new model object. Integrity Error: {e}"
                    )
                except AttributeError as e:
                    logger.info(
                        f"Could not create new model object. AttributeError. Error: {e}"
                    )
                    self.debug_message(
                        f"\t\t Could not create new model object. AttributeError. Error: {e}"
                    )
                except Exception as e:
                    logger.error(
                        f"Unhandled error. Could not create a new object for {model._meta.verbose_name}. Error: {e}"
                    )
                    self.debug_message(
                        f"\t\t Unhandled error. Could not create a new object for {model._meta.verbose_name}. Error: {e}"
                    )

        return self.created, self.skipped, self.updated


class Command(BaseCommand):
    help = "Import data from an Airtable and overwrite model or page information"

    def add_arguments(self, parser):
        parser.add_argument(
            "labels",
            metavar="model_name",
            nargs="+",
            help="Model (as app_label.model_name) or app name to populate table entries for, e.g. creditcards.CreditCard",
        )

    def handle(self, *args, **options):
        """
        Runs the management command with the app_name.ModelName as parameters.
        ie. python manage.py import_airtable pages.HomePage creditcards.CreditCard

        This will handle Wagtail Pages (with page revisions) along with standard
        Django models.

        Models passed in to the command go through a quick validation to ensure they
        exist.

        Every model is then looped through, and a `settings.AIRTABLE_IMPORT_SETTINGS`
        is searched for based on the model's label (ie. pages.HomePage). These
        settings are used to connect to a certain Airtable Base (a set of spreadsheets),
        the name of the table to use, a unique identifier (used for connecting previously
        unrelated Airtable records to Django objects), and a serializer for validating
        incoming data from Airtable to make it work with the Django field types.

        Each model loop contains an `airtable.get_all()` command which
        will get all the data from the Airtable Table and load it into a
        list of dictionaries. This uses memoization to reduce memory hogging and
        wasted API calls to Airtable.

        Then every record is iterated over and 3 actions are taken:
            1. Look for an existing model object by its airtable_record_id.
               Update if found. Skip to step 2 if not found.
            2. Search for a model object by its unique identifier.
               If found, update the object. Skip to step 3 if not found.
            3. Create a new object. This step is very susceptible to fail based on the
               model type and fields in the serializer. Wagtail pages cannot be created
               from Airtable records as there's missing data such as depth and parent
               pages. And so Wagtail pages are skipped entirely in step 3.
        """
        # Overwrite verbosity if WAGTAIL_AIRTABLE_DEBUG is enabled.
        if settings.DEBUG:
            # AIRTABLE_DEBUG can only be enabled if standard Django DEBUG is enabled.
            # The idea is to protect logs and output in production from the noise of these imports.
            AIRTABLE_DEBUG = getattr(settings, "WAGTAIL_AIRTABLE_DEBUG", False)
            if AIRTABLE_DEBUG:
                options["verbosity"] = 2

        importer = Importer(models=options["labels"], options=options)
        created, skipped, updated = importer.run()

        if options["verbosity"] >= 1:
            self.stdout.write(
                f"{created} object created. {updated} object updated. {skipped} object skipped."
            )

        # Use friendlier message. Slightly differs from the stdout.write() string.
        return f"{created} items created. {updated} items updated. {skipped} items skipped."



#!/usr/bin/env python

from setuptools import setup, find_packages

install_requires = [
    "wagtail>=2.6",
    "airtable-python-wrapper>=0.13.0",
    "djangorestframework>=3.11.0,<=3.12.2",
]

setup(
    name='wagtail-airtable',
    version='0.1.6',
    description="Sync data between Wagtail and Airtable",
    author='Kalob Taulien',
    author_email='kalob.taulien@torchbox.com',
    url='https://github.com/wagtail/wagtail-airtable',
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    license='BSD',
    long_description="An extension for Wagtail allowing content to be transferred between Airtable sheets and your Wagtail/Django models",
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Django',
        'Framework :: Wagtail',
        'Framework :: Wagtail :: 2',
    ],
    install_requires=install_requires,
)



"""
Utility functions for wagtail-airtable.
"""
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist

from wagtail_airtable.mixins import AirtableMixin

from wagtail.admin import messages

WATAIL_AIRTABLE_ENABLED = getattr(settings, "WAGTAIL_AIRTABLE_ENABLED", False)


def get_model_for_path(model_path):
    """
    Given an 'app_name.model_name' string, return the model class or False
    """
    app_label, model_name = model_path.lower().split(".")
    try:
        return ContentType.objects.get_by_natural_key(
            app_label, model_name
        ).model_class()
    except ObjectDoesNotExist:
        return False


def get_all_models() -> list:
    """
    Gets all models from settings.AIRTABLE_IMPORT_SETTINGS.

    Returns a list of models.
    """
    airtable_settings = getattr(settings, "AIRTABLE_IMPORT_SETTINGS", {})
    validated_models = []
    for label, model_settings in airtable_settings.items():
        if model_settings.get("AIRTABLE_IMPORT_ALLOWED", True):
            label = label.lower()
            if "." in label:
                try:
                    model = get_model_for_path(label)
                    validated_models.append(model)
                except ObjectDoesNotExist:
                    raise ImproperlyConfigured(
                        "%r is not recognised as a model name." % label
                    )

    return validated_models


def get_validated_models(models=[]) -> list:
    """
    Accept a list of model paths (ie. ['appname.Model1', 'appname.Model2']).

    Looks for models from a string and checks if the mode actually exists.
    Then it'll loop through each model and check if it's allowed to be imported.

    Returns a list of validated models.
    """
    validated_models = []
    for label in models:
        if "." in label:
            # interpret as a model
            model = get_model_for_path(label)
            if not model:
                raise ImproperlyConfigured(
                    "%r is not recognised as a model name." % label
                )

            validated_models.append(model)

    models = validated_models[:]
    for model in validated_models:
        airtable_settings = settings.AIRTABLE_IMPORT_SETTINGS.get(model._meta.label, {})
        # Remove this model from the `models` list so it doesn't hit the Airtable API.
        if not airtable_settings.get("AIRTABLE_IMPORT_ALLOWED", True):
            models.remove(model)

    return models


def can_send_airtable_messages(instance) -> bool:
    """
    Check if a model instance is a subclass of AirtableMixin and if it's enabled.
    """

    # Check if the page is an AirtableMixin Subclass
    # When AirtableMixin.save() is called..
    # Either it'll connect with Airtable and update the row as expected, or
    # it will have some type of error.
    # If _airtable_update_error exists on the page, use that string as the
    # message error.
    # Otherwise assume a successful update happened on the Airtable row
    if (
        WATAIL_AIRTABLE_ENABLED and issubclass(instance.__class__, AirtableMixin)
        and hasattr(instance, "is_airtable_enabled") and instance.is_airtable_enabled
    ):
        return True
    return False


def airtable_message(request, instance, message="Airtable record updated", button_text="View record in Airtable", buttons_enabled=True) -> None:
    """
    Common message handler for Wagtail hooks.

    Supports a custom message, custom button text, and the ability to disable buttons entirely (use case: deleting a record)
    """
    if hasattr(instance, "_airtable_update_error"):
        messages.error(request, message=instance._airtable_update_error)
    else:
        buttons = None
        if buttons_enabled and instance.get_record_usage_url():
            buttons = [
                messages.button(instance.get_record_usage_url(), button_text, True)
            ]
        messages.success(request, message=message, buttons=buttons)





from rest_framework import serializers


class AirtableSerializer(serializers.Serializer):
    """
    Generic Airtable serializer for parsing Airtable API JSON data into proper model data.
    """

    def validate(self, data):
        """
        Loop through all the values, and if anything comes back as 'None' return an empty string.

        Not all fields will have cleaned data. Some fields could be stored as 'None' in Airtable,
        so we need to loop through every value and converting 'None' to ''
        """
        for key, value in data.items():
            # If any fields pass validation with the string 'None', return a blank string
            if value == "None":
                data[key] = ""
        return data



from django import forms
from django.conf import settings


class AirtableImportModelForm(forms.Form):

    model = forms.CharField()

    def clean_model(self):
        """Make sure this model is in the AIRTABLE_IMPORT_SETTINGS config."""

        model_label = self.cleaned_data["model"].lower()
        airtable_settings = getattr(settings, "AIRTABLE_IMPORT_SETTINGS", {})
        is_valid_model = False

        for label, model_settings in airtable_settings.items():
            if model_label == label.lower():
                is_valid_model = True
                break

        if not is_valid_model:
            raise forms.ValidationError("You are importing an unsupported model")

        return model_label
